'use strict';

exports.__esModule = true;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Reduxable = function () {
  /*
  *  In the constructor we set a new method for each reducer.
  *
  *  When called, each method will dispatch an action with the convenient `type` and `scope`
  *  so the reducer will catch this action.
  *
  *  ```
  *  class ReduxableExample extends Reduxable {
  *    reducers = {
  *      foo: () => {}
  *    }
  *  }
  *
  *  const reduxableInstance = new ReduxableExample()
  *  reduxableInstance.foo('hello world')
  *  // Will dispatch an action like => { type: 'foo', payload: 'hello world', scope: '...' }
  *  ```
  */

  function Reduxable() {
    var _this = this;

    _classCallCheck(this, Reduxable);

    this.state = {};

    var _loop = function _loop(reducerName) {
      var actionForReducer = _this.actions[reducerName];
      _this.state[reducerName] = function (payload) {
        return _this.dispatch(actionForReducer(payload));
      };
    };

    for (var reducerName in this.actions) {
      _loop(reducerName);
    }
  }

  /*
  *  The `_store` will be the Redux store. Since this store is unique by application, then we
  *  can save _statically_ and use it across all the Reduxable instances.
  *
  *  We use this store internally on two Reduxable instance methods:
  *    - `dispatch` to precisely dispatch the actions
    this.constructor.showWarnings = true
  *    - `getState` to retrieve the portion of state corredpondent to the Reduxable instance
  *
  *  This method is called from `createStore` method. See its documentation for more details.
  */

  Reduxable.setStore = function setStore(store) {
    this._store = store;
  };

  /*
  *  The `_scope` will define where this Reduxable instance is placed on the global state tree
  *
  *  This scope will be set as a parameter in each dispacthed action and will be used to determine
  *  whether or not an action should be catched by a reducer.
  *
  *  This method is called from `combineReducers` method. See its documentation for more details.
  */

  Reduxable.prototype.setScope = function setScope(scope) {
    this._scope = scope;
  };

  /*
  *  If store has been already set, then call the store's `dispatch`
  *  If not, apply the reducer to the current state and set it in `_localState`
  */


  Reduxable.prototype.dispatch = function dispatch(action) {
    var store = this.constructor._store;
    if (store) {
      return store.dispatch(action);
    }

    var type = action.type;
    var payload = action.payload;

    this._localState = this.constructor.reducers[type](this.getState(), payload);
  };

  /*
  *  Returns the state for this particular scope
  *
  *  Given the following structure for a Redux store
  *
  *  const reduxStore = createStore(combineReducers({
  *    a: combineReducers({
  *      b: thisReduxableInstance
  *    })
  *  })
  *
  *  Then `thisReduxableInstance.getState()` will return `reduxStore.getState().a.b`
  */


  Reduxable.prototype.getState = function getState() {
    if (!this.constructor._store) {
      return this._localState || this.initialState;
    }

    var state = this.constructor._store.getState();
    if (!this._scope) {
      return state;
    }

    return this._scope.split('.').reduce(function (object, scopeKey) {
      return object[scopeKey];
    }, state);
  };

  /*
  *  Returns a reducer function grouping all the defined reducers
  *
  *  This method will do the following
  *    1) Check that the action `scope` match with the scope of this Reduxable instance
  *    2) Find the method that matchs with the action `type`
  *    3) Call that method with the current state and the action `payload`
  *    4) Check that the new state retrieved by that method is not the same
  *       (i.e the method did not mutate the previous state)
  */


  Reduxable.prototype.getReducer = function getReducer() {
    var _this2 = this;

    var warn = this.constructor.showWarnings && console ? console.warn : function () {};

    return function () {
      var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : _this2.initialState;
      var _ref = arguments[1];
      var type = _ref.type;
      var scope = _ref.scope;
      var payload = _ref.payload;

      if (scope !== _this2._scope) {
        return state;
      }

      var method = _this2.constructor.reducers[type];

      if (method) {
        var newState = method(state, payload);
        if (typeof state === 'object' && state === newState) {
          warn('Reducer \'' + type + '\' in scope \'' + scope + '\' is returning the same object.\n            If you are using Immutable this is nothing to worry about.\n            You can disable this with \'Reduxable.showWarnings = false\'');
        }
        return newState;
      }

      return state;
    };
  };

  /*
  *  Returns an object with all the action creators for the defined reducers
  *
  *  This method will iterate over all the defined `reducers` and create a new
  *  object of methods (action creators)
  *
  *  ```
  *  class ReduxableExample extends Reduxable {
  *    reducers = {
  *      foo: () => {}
  *    }
  *  }
  *
  *  const reduxableInstance = new ReduxableExample()
  *  const action = reduxableInstance.actions.foo('hello world')
  *  console.log(action)
  *  // { type: 'foo', payload: 'hello world', scope: '...' }
  *  ```
  */


  _createClass(Reduxable, [{
    key: 'actions',
    get: function get() {
      var _this3 = this;

      if (!this._actions) {
        this._actions = {};

        var _loop2 = function _loop2(reducerName) {
          if (_this3.constructor.reducers.hasOwnProperty(reducerName)) {
            _this3._actions[reducerName] = function (payload) {
              return { payload: payload, type: reducerName, scope: _this3._scope };
            };
          }
        };

        for (var reducerName in this.constructor.reducers) {
          _loop2(reducerName);
        }
      }

      return this._actions;
    }
  }]);

  return Reduxable;
}();

Reduxable.showWarnings = true;

exports['default'] = Reduxable;